#include "sim.h"
#include <stdio.h>
#include <sys/termios.h>
#include <signal.h>

int trapNo= -1;  /* No trap */
int trapAddr;
int gotInt;

int Usr1= 0;
int HandleUsr1()
{
 Usr1= 1;
}

int handleTrap()
     /* Return of 0 indicates stop (return to X interface), 1 => go on */
{
 extern char radix;
 int goOn= gotInt ? 0 : 1;

 if (trapNo != -1)
 {
  switch (trapNo & TRAP_MASK)
  {
   case HALT_TRAP:
    message("Processor halted");
    pc= oldpc;
    goOn= 0;  /* Return to X interface */
    break;

   case NO_MEM_TRAP:
    pc= oldpc;
    message("Attempted to %s non-existent memory location %s",
	    (trapNo & MEM_READ_FLAG) ? "read" :
	     ((trapNo & MEM_WRITE_FLAG) ? "write" :
	      ((trapNo & MEM_FETCH_FLAG) ? "execute at" : "<unknown>")),
	    NumOutNoPad(trapAddr, 1, 0xffffffff, radix));
    goOn= 0;
    break;
   case MISALIGNED_TRAP:
    pc= oldpc;
    message("Attempted to %s misaligned memory location %s",
	    (trapNo & MEM_READ_FLAG) ? "read" :
	    ((trapNo & MEM_WRITE_FLAG) ? "write" :
	     ((trapNo & MEM_FETCH_FLAG) ? "execute at" : "<unknown>")),
	    NumOutNoPad(trapAddr, 1, 0xffffffff, radix));
    goOn= 0;
    break;

   case IBPT_TRAP: case RBPT_TRAP: case WBPT_TRAP:
    message("%s breakpoint reached", (trapNo == IBPT_TRAP ? "Fetch" :
				      (trapNo == RBPT_TRAP ? "Read" :
				       "Write")));
    pc= oldpc;
    goOn= 0;
    break;
   case KBD_READ_TRAP:
   {
    /* We check for gotInt here, which is used for both single-stepping
       and interrupts.  If it's a single-step, ignoreBpt is set, so we
       set gotInt to 0, and restore it afterwards. */
    int old_gotInt= gotInt;
    if (ignoreBpt) gotInt= 0;
    if (!using_x)
    {
     int old_lflag, old_vmin, old_vtime, old_iflag;
     int tmp;
     static struct termios tio;
     if (ioctl(0, TCGETS, &tio) == 0)
     {
      old_lflag= tio.c_lflag;
      old_iflag= tio.c_iflag;
      old_vmin= tio.c_cc[VMIN];
      old_vtime= tio.c_cc[VTIME];
      tio.c_lflag &= ~ICANON;
      tio.c_lflag &= ~ECHO;
      tio.c_lflag |= ISIG;
      tio.c_iflag &= ~ICRNL;
      tio.c_cc[VMIN]= 1;
      tio.c_cc[VTIME]= 0;  /* Wait for one character. */
      ioctl(0, TCSETS, &tio);
     }
     for (;;)
     {
      extern int errno;
      struct sigaction ovec, vec;
      sigaction(SIGINT, NULL, &ovec);
      vec= ovec;
      vec.sa_flags &= ~SA_RESTART;  /* So getchar() exits when user presses
				      control-c */
      sigaction(SIGINT, &vec, NULL);
      errno= 0;
      tmp= getchar();
      if (!is_tty && tmp == '\n') tmp= '\r';
      sigaction(SIGINT, &ovec, NULL);
      if (tmp == -1)  /* Probably interrupted */
      {
       if (!gotInt)  /* An error.  Stop anyway */
       {
	gotInt= 1;
	if (errno)
	 perror("read");
       }
       goOn= 0;
       if (!gotInt) gotInt= old_gotInt;
       pc= oldpc;
       break;
      }
      if (tmp < 128 && tmp >= 0)
      {
       regs[1]= (unsigned char)tmp;   /* Endian */
       if (!gotInt) gotInt= old_gotInt;
       break;
      }
      else
       write(1, "\007", 1);
     }  /* for */
     tio.c_lflag= old_lflag;
     tio.c_iflag= old_iflag;
     tio.c_cc[VMIN]= old_vmin;
     tio.c_cc[VTIME]= old_vtime;
     ioctl(0, TCSETS, &tio);
    }
#ifndef SIM_MAIN
    else
    {
     if (Usr1)
     {
      message("PANIC: Got SIGUSR1 without asking for it!\n");
      Usr1= 0;
     }
     if (gotInt && !ignoreBpt) return 1;
     write(comms_sock[1], "i", 1);
     for (;;)
     {
      pause();
      if (Usr1)
      {
       goOn= old_gotInt ? 0 : 1;  /* Don't go on if we got an int. or if
				     we're single-stepping */
       Usr1= 0;
       if (!gotInt) gotInt= old_gotInt;
       break;
      }
      else if (gotInt)
      {
       goOn= 0;
       pc= oldpc;
       if (!gotInt) gotInt= old_gotInt;
       break;
      }
     } /* for */
    } /* using_x */
#endif
   }
   break;
   case SCREEN_WRITE_TRAP:
    /*printf("SCREEN WRITE HERE\n");*/
    if (!using_x)
    {
     int old_oflag;
     int tmp;
     static struct termios tio;
     tmp= regs[1];
     if ((tmp >= ' ' && tmp <= '~') || tmp == '\n' || tmp == '\r' ||
	 tmp == '\007')
     {
      if (ioctl(0, TCGETS, &tio) == 0)
      {
       old_oflag= tio.c_oflag;
       tio.c_lflag &= ~ONLCR;
       ioctl(0, TCSETS, &tio);
      }
      putchar((char)regs[1]);  /* Endian */
      tio.c_oflag= old_oflag;
      ioctl(0, TCSETS, &tio);
     }  /* Char in range */
    }
#ifndef SIM_MAIN
    else
    {
     char tmp= regs[1];
     write(comms_sock[1], "o", 1);
     write(comms_sock[1], &tmp, 1);
    }
#endif
    break;

   case OVERFLOW_TRAP:
    message("Integer overflow");
    pc= oldpc;
    goOn= 0;
    break;

   case UNIMP_TRAP:
    message("Unimplemented unstruction");
    pc= oldpc;
    goOn= 0;
    break;

   case FP_TRAP:
    message("Floating point operation attempted");
    goOn= 0;
    break;
 
   case PANIC_TRAP:
    /* Message has already been printed */
    pc= oldpc;
    goOn= 0;
    break;
    
   default:
    message("Unimplemented trap %d", trapNo);
    pc= oldpc;
    goOn= 0;
    break;
  }
 } /* trapNo != -1 */
 if (gotInt && goOn) goOn= 0;
 gotInt= 0;
 trapNo= -1;
 trapAddr= -1;  /* For error checking */
 return goOn;
}

